(*i camlp4deps: "parsing/grammar.cma" i*)
(* 
Arbre de preuve courant.

la fonction preuve_courante () rend l'arbre de preuve courant dans le type preuve.

*)

open Clenv
open Command
open Coqast
open Decl_kinds
open Declarations
open Declare
open Eauto
open Entries
open Environ
open Equality
open Esyntax
open Evarutil
open Evd
open Extend
open Genarg
open Inductive
open Inductiveops
open Libnames
open Metasyntax
open Nameops
open Names
open Pcoq
open Pfedit
open Pp
open Printer
open Proof_trees
open Proof_type
open Rawterm
open Reduction
open Reductionops
open Refiner
open Sign
open Tacexpr
open Tacmach
open Tacred
open Tacticals
open Tactics
open Term
open Termops
open Typing
open Util
open Vernacentries
open Vernacinterp

(***********************************************************************
   Le type des arbres de preuves 
*)

type preuve =
    (* but sans preuve *)
  |Taprouver of
      preuve * (* contexte nouveau: ne contient que les hypotheses
		  qui sont nouvelles par rapport au contexte du but pere*)
      preuve (* but *)
	(* but avec preuve *)
  |Tpreuve of
      preuve * (* contexte nouveau *)
      preuve * (* but *)
      preuve * (* tactique *)
      (preuve list) (* preuves des sous-buts *)
	(* tactique *)
  |Ttactic of
      (rule * int) * (* tactique et son rang dans le script *)
      (rule * int) list * (* utilise par l'explication en langue naturelle *)
      (rule * int) list  (* utilise par l'explication en langue naturelle *)
      
      (* nom d'hypothese *)
  |Tid of string
      (* terme des constructions *)
  |Tterm of env * constr (* l' environnement et le terme *)
      (* hypothese du contexte *)
  |Thyp of
      preuve * (* le nom de l'hypothese *)
      preuve * (* sa valeur eventuelle: un terme, ou Tvide *)
      preuve (* son type *)
	(* contexte nouveau d'un but *)
  |Tcontexte of preuve list (* liste d' hypotheses *)
  |Tvide 


(***********************************************************************
   Contextes, buts
*)
exception Different

(* enleve de signe ce qui se trouve dans osign *)
let new_sign osign sign =
  Sign.fold_named_context
    (fun (id,c,ty as d) sign ->
       try 
	 if Sign.lookup_named id osign = (id,c,ty) then sign
	 else raise Different
       with Not_found | Different -> add_named_decl d sign)
    sign ~init:empty_named_context

(* change les hypothese d'un but *)
let change_hyps hyps pf=
  let {evar_hyps=_; evar_concl=cl; evar_body=body} = pf.goal in
    {evar_hyps=hyps; evar_concl=cl; evar_body=body}
    
(* hypotheses d'un but *)
let hyps pf =
  let {evar_hyps=hyps; evar_concl=cl; 
       evar_body=body} = pf.goal in
    hyps

(* conclusion d'un but *)
let concl pf = 
 let {evar_hyps=hyps; evar_concl=cl; 
       evar_body=body} = pf.goal in
   cl

(* environnement d'un  but *)
let env pf =(evar_env pf.goal)

(* sous-buts d'un but *)
let sous_buts pf =
  match pf.ref with
    | None -> []
    | Some(r,lpf) -> lpf

(* tactique d'un but *)
let tactic pf =
  match pf.ref with
    | None -> assert false
    | Some(r,lpf) -> r

(* expression vernac d'une tactique *)
let string_of_rule r =
  pp_with  Format.str_formatter (pr_rule r);
  let s = Format.flush_str_formatter () in
    s^"."


let rec string_of_preuve p = match p with
    |Taprouver (p1,p2) -> "aprouver"
  |Tpreuve (c,b,t,lb)->
      (List.fold_left (fun s b -> s^"\n"^(string_of_preuve b))
          ((string_of_preuve c)^"|-"^(string_of_preuve b)^" by "^(string_of_preuve t)^"=>") lb)^"\n FIN\n"
  |Ttactic ((r,_),_,_)-> string_of_rule r
  |Tid s -> s
  |Tterm (e,c) -> "TERME"
  |Thyp (h,v,t)-> string_of_preuve h
  |Tcontexte lp -> List.fold_left (fun s b -> s^", "^(string_of_preuve b))
          "" lp
  |Tvide -> ""

let print_preuve p =
  print_string ((string_of_preuve p)^"\n")

(***********************************************************************
   Expansion des tactique(lle)s
*)

(* est-ce-qu'on doit developper la tactique? *)
let rule_is_complex r = 
   match r with
     | Tactic (TacArg (Tacexp (TacArg (TacCall _))),_)-> false
     | Tactic (TacArg (Tacexp (TacArg (Reference _))),_)-> false
     | Tactic (TacArg (Tacexp (TacThens _)),_)-> true
     | Tactic (TacArg (Tacexp (TacThen _)),_)-> true
     | Tactic (TacArg (Tacexp  _),_)-> true
     | Tactic (TacAtom (_,TacAuto _), _) -> false
     |_ -> false


(* Attribue les preuves de la liste l aux sous-buts non-prouvs de pf  *)
let remplit l pf =
  let lpf = ref l in
  let rec remplit pf =
    match pf.ref with
        None -> let p=List.hd (!lpf) in
	  lpf:=List.tl (!lpf);
	  p
      | Some(r,lspf)->
	  let l1=List.map remplit lspf in
	    {open_subgoals = pf.open_subgoals;
	     goal = pf.goal;
	     ref = Some(r,l1)}
  in remplit pf

let rec expanser_tactiquelles expanser pf =
  match pf.ref with
    | None -> pf
    | Some(r,lpf) ->
	(match r with
	   | Tactic (TacAtom (_,TacAuto _),pf1)
	   | Tactic (TacArg (Tacexp _),pf1)->
	       if expanser && (rule_is_complex r)
	       then expanser_tactiquelles expanser (remplit lpf pf1)
	       else {open_subgoals = pf.open_subgoals;
		     goal = pf.goal;
		     ref = Some(r,List.map (expanser_tactiquelles expanser) lpf)}
	   |_ -> {open_subgoals = pf.open_subgoals;
		  goal = pf.goal;
		  ref = Some(r,List.map (expanser_tactiquelles expanser) lpf)}
	)

(***********************************************************************
   Traduction d'un arbre de preuve Coq dans le type preuve
*)

(* focntion principale *)
let rec preuve0 pf =
  preuve empty_named_context pf
    
(* fonction recursive *)
and preuve osign pf =
  let hyps0 = (hyps pf) in
  let cl0=(concl pf) in
  let hyps' = new_sign osign hyps0 in (* les hypotheses nouvelles *)
    match pf.ref with
      | None -> traduit_a_prouver hyps0 (change_hyps hyps' pf) pf
      | Some(r,spfl) -> traduit_sous_buts hyps0 (change_hyps hyps' pf) pf r spfl

and traduit_a_prouver hyps0 hyps pf =
  Taprouver (arbre_contexte (env pf) hyps,Tterm (env pf,concl pf))

and traduit_sous_buts hyps0 hyps pf r lpf =
  Tpreuve (arbre_contexte (env pf) hyps, Tterm (env pf,concl pf),
	   Ttactic ((r,0),[],[]),
	   List.map (preuve hyps0) lpf)
      
and arbre_contexte env0 hyps =
  let lterm=ref [] in
  let l=Environ.fold_named_context
	  (fun env (id,c,ty) a ->
	     let c = match c with None -> Tvide | Some c -> Tterm (env0,c) in
	     let t = Thyp (traduit_id id, c  , Tterm (env0,ty))
	     in
	       lterm:=(!lterm)@[t];
	       a)
          (evar_env hyps) ~init:(mt ()) in
   Tcontexte (!lterm)

and traduit_id id = Tid (string_of_id id)




(* arbre de preuve courant en format Coq *)
let get_pf () =
  let pf = (Tacmach.proof_of_pftreestate (Pfedit.get_pftreestate())) in
    pf

(* Met les rangs des tactiques dans le script *)

let met_rangs p =
  let i = ref 0 in
  let rec met p =
    match p with
      |Tpreuve (h,c,Ttactic ((r,_),_,_),lp) ->
	  i:=(!i)+1;
	  let j=(!i) in
	  let lp = List.map met lp in
	  Tpreuve (h,c,Ttactic ((r,j),[],[]),lp)
      |_->p
  in met p
    
(***********************************************************************
  Fonction principale: rend l'arbre de preuve courant dans le type preuve
***********************************************************************)

let preuve_courante expanser =
  met_rangs (preuve0 (expanser_tactiquelles expanser (get_pf ())))
  
